#include <pthread.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include "structs.h"
#include "core.h"
#include "threads.h"
#include "ocsp_codes.h"
#include "ocsp_request.h"
#include "ocsp_response.h"
#include "ocspd_pkey.h"

extern OCSPD_CONFIG *ocspd_conf;

extern pthread_mutex_t clifd_mutex;
extern pthread_cond_t clifd_cond;

extern pthread_mutex_t threads_mutex;
extern pthread_cond_t threads_cond;

extern pthread_mutex_t crl_mutex;

pthread_mutex_t resp_mutex = PTHREAD_MUTEX_INITIALIZER;

int thread_make ( int i ) {
	int ret;

	if ((ret = pthread_create(&ocspd_conf->threads_list[i].thread_tid, 
			NULL, thread_main, (void *) (ocspd_conf->clifd + i)))) {

		syslog(LOG_ERR,"ERROR::OPENCA_SRV_ERR_THREAD_CREATE");
		return(ret);
	}

	return(OCSPD_SRV_OK);
}

void * thread_main ( void *arg ) {
	int connfd = -1;

	OCSP_REQUEST  *req = NULL;
	OCSP_RESPONSE *resp = NULL;
	EVP_PKEY *pkey;
	ENGINE *engine;
	int thread_nr;

	struct sigaction sa;

	thread_nr = (int *) arg - ocspd_conf->clifd;

	if ( ocspd_conf->verbose ) {
		syslog(LOG_INFO, "INFO::OPENCA_SRV_INFO_TREAD::new thread "
			"created [%d]", thread_nr);
	}

	sa.sa_handler = SIG_IGN;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;

	if (sigaction(SIGPIPE, &sa, NULL) == -1) {
		perror("Error during handling the death processes");
		exit(1);
	}

#ifdef HAVE_ENGINE
	/*
	if(ocspd_conf->use_engine == 1) {
		syslog(LOG_INFO, "INFO::Initializing ENGINE" );
		if((engine = ocspd_init_engine( ocspd_conf )) == NULL ) {
			syslog (LOG_ERR, "[Thread::%d] Can not initialize "
				"the ENGINE interface, aborting!",
				thread_nr );
			pthread_exit( NULL );
		}
	}
	*/

	if ( ocspd_conf->use_engine == 1 ) {
		engine = ocspd_conf->engine;
	}
#else
	engine = NULL;
#endif
	/*
	if( ocspd_conf->verbose == 1 ) {
		syslog( LOG_INFO, "INFO::Loading private key" );
	}

	if( (pkey = ocspd_load_key ( ocspd_conf, engine )) == NULL ) {
		syslog(LOG_ERR, "[Thread::%d] Can not load Key!",
						thread_nr );
		syslog(LOG_ERR, "Fatal Error, ABORTING server.");
		kill( getpid(), SIGHUP );

		// syslog(LOG_ERR, "ERROR::Thread PID is %d", getpid());
		pthread_exit(NULL);
	}
	*/

	if( ocspd_conf->pkey == NULL ) {
		syslog(LOG_ERR, "[Thread::%d] Private Key not loaded, Aborting",
						thread_nr );
		kill( getpid(), SIGHUP );
		pthread_exit ( NULL );
	}

	pkey = ocspd_conf->pkey;

	/*
	if( ocspd_conf->verbose == 1 ) {
		syslog( LOG_INFO, "[Thread::%d] Private Key loaded "
					"successfully (%p)", thread_nr, pkey );
	}
	*/

	for ( ; ; ) {

		/* Before calling the cond_wait we need to own the mutex */
		pthread_mutex_lock( &threads_mutex );

		if( ocspd_conf->debug == 1 ) {
			syslog( LOG_INFO, "DEBUG::Thread [%d] waiting on "
				"clifd_cond", thread_nr );
		}

		/* Wait on the condition variable */
		pthread_cond_wait( &threads_cond, &threads_mutex );

		if( ocspd_conf->debug == 1 ) {
			syslog( LOG_INFO, "DEBUG::Thread [%d] woke up on "
					"clifd_cond", thread_nr );
		}

		/* Lock the mutex to be able to read the content safely */
		// pthread_mutex_lock( &clifd_mutex );
		/*
		while ( ocspd_conf->iget == ocspd_conf->iput ) {
			if( ocspd_conf->verbose == 1 ) {
				syslog( LOG_INFO, "DEBUG::Thread [%ld] cond "
					"wait clifd_cond, clifd_mutex", 
						(long) pthread_self() );
			}
			pthread_cond_wait( &clifd_cond, &clifd_mutex );
			if( ocspd_conf->debug ) {
				fprintf(stderr, 
					"[Thread] - awaken (iget=%d, iput=%d)\n",
					ocspd_conf->iget, ocspd_conf->iput );
			}
		}
		*/

		/* Get the his own spot, check if there is a valid fd */
		while ( ocspd_conf->clifd[thread_nr] == -1 ) {
			pthread_cond_wait( &threads_cond, &threads_mutex );
		}
		pthread_mutex_unlock( &threads_mutex );

		// pthread_mutex_lock( &clifd_mutex );
		connfd = ocspd_conf->clifd[thread_nr];
		// pthread_mutex_unlock( &clifd_mutex );

		if( connfd < 0 ) {
			/* If its own spot is not a valid fd, go from the
 			 * beginning */
			if( ocspd_conf->verbose == 1 ) {
				syslog( LOG_INFO, "[Thread::%d] Back to sleep ",
					thread_nr);
			}

			continue;
		}

		// connfd = ocspd_conf->clifd[ocspd_conf->iget];
		if( ocspd_conf->verbose )  {
			syslog(LOG_INFO, "[Thread::%d] got connd %d", 
							thread_nr, connfd);
		}
		/*
		fprintf( stderr, "TREAD::%d got connfd=%d from slot %d!\n",
				(int) arg, connfd, ocspd_conf->iget);
		*/

		// ocspd_conf->iget += 1;
		// if( ocspd_conf->iget == ocspd_conf->nthreads ) 
		// 	ocspd_conf->iget = 0;

		// pthread_mutex_unlock( &clifd_mutex );
		// fprintf(stderr, "[Thread] - released clifd_mutex\n");

		// Why do we need this ???
		// ocspd_conf->threads_list[ (int) arg ].thread_count++;


		if((req = ocspd_req_get_socket(connfd, ocspd_conf)) == NULL ) {
			syslog(LOG_ERR, "ERROR::Can not parse REQ");
			goto err;
		}

		if( ocspd_conf->debug ) {
			syslog(LOG_INFO, "[Thread::%d] Read response from "
						"socket", thread_nr );
		}

		if((resp = make_ocsp_response(req, ocspd_conf, pkey)) == NULL){
                        syslog( LOG_ERR, "ERROR::Can not generate response!" );
			pthread_mutex_unlock( &resp_mutex );
                        goto err;
		}

		if( ocspd_conf->debug) {
			syslog(LOG_INFO, "[Thread::%d] Built response from "
						"socket", thread_nr);
		}

		goto end;

err:
		if( ocspd_conf->debug) {
			syslog(LOG_ERR, "[Thread::%d] - An error occurred!",
					thread_nr);
		}

		if((resp = OCSP_response_create( 
			OCSP_RESPONSE_STATUS_MALFORMEDREQUEST, NULL)) == NULL) {

			if( req ) OCSP_REQUEST_free (req);
			req = NULL;

                	ocspd_resp_send_socket( connfd, resp, ocspd_conf );
			close( connfd );

			pthread_mutex_lock( &clifd_mutex );
			ocspd_conf->clifd[thread_nr] = -1;
			pthread_mutex_unlock( &clifd_mutex );

			continue;
		}

end:
                ocspd_resp_send_socket( connfd, resp, ocspd_conf );

		pthread_mutex_lock( &resp_mutex );
		if( resp ) {
			OCSP_RESPONSE_free (resp);
			resp = NULL;
		}
		pthread_mutex_unlock( &resp_mutex );

		if( req ) {
			OCSP_REQUEST_free (req);
			req = NULL;
		}

		close( connfd );

		/* Set the connfd[thread_nr] to -1 indicating that the thread
 		 * is actually free and can take another job */
		pthread_mutex_lock( &clifd_mutex );
		ocspd_conf->clifd[thread_nr] = -1;
		pthread_mutex_unlock( &clifd_mutex );
	}
}
